# Laravel Event Driven Service

## Install
### Add Myriad Repository

The client is distributed internally with Satis. First add the repository to `composer.json`:

```json
  "repositories":  [
    {
      "type": "composer",
      "url": "https://satis.myriadmobile.com"
    }
  ]
```

### Require the Dependency

Require this package with composer using the following command:

```bash
composer require myriad/event-driven-service
```

### Install Service Provider

Add the event driven service provider to the `providers` array in the `config/app.php`

```php
Myriad\EventDrivenService\EventDrivenServiceProvider::class,
```

### Publish Migrations

Rather than creating your own migrations table, you can publish the one from the event driven service package:

```bash
php artisan vendor:publish --provider="Myriad\EventDrivenService\EventDrivenServiceProvider" --tag="migrations"
```

### Configuration

The event driven service package is designed to work with a configuration. Publish the event driven service config file to the `config` directory:

```bash
php artisan vendor:publish --provider="Myriad\EventDrivenService\EventDrivenServiceProvider" --tag="config"
```

The config file is where you can define the models that will utilize the event driven system.

The `uri` key points to the name of the `uri` that will be passed in requests.

The `command` key points to the name of the model that will passed in the `command` value of the event request.

The `model` key points to the model class that will be receiving the command and operation.

Example:
```php
'models' => [
  [
    'uri' => 'carriers',
    'command' => 'carrier',
    'model' => \App\Models\Carrier::class
  ],
]
```

### Routes

There are two routes that the package utilizes. Add the following lines to your `api.php` route files.

```php
Route::apiResource('events', '\Myriad\EventDrivenService\Http\Controllers\Api\EventsController')->only(['store']);
Route::get('events/{uri}', '\Myriad\EventDrivenService\Http\Controllers\Api\EventsController@getEvents');
```

### Users

Laravel's default `User` model will need to be adjusted to work correctly with this package. 

#### Migrations
The `User` migration will use UUIDs instead of IDs. Be sure to publish the migrations included in this package to get the updated `User` migration.

#### User Model

##### Properties
The `User` model will need to have two properties defined. Add or replace the following two lines to your `User` model:

```php
public $incrementing = false;
protected $primaryKey = 'uuid';
```

##### Methods
Please add or replace the following method to the `User` model:
```php
protected static function boot() {
  parent::boot();
  static::creating(function (User $user) {
    $user->uuid = Uuid::uuid4();
  });
}
```

## Usage
### How it works

The event driven service package utilizes models that extend off the base `EventModel` included in the package. 

The models can implement the `IUserFilterable` interface, which included a query scope `scopeFilterForUser($query)`, which can be overridden to customize the query results to an authenticated user.

Clients will `POST` "event" requests to the server to create, update or delete models. An `EventService` will handle the `POST` requests and perform the intended operation (create, update, delete) on the model if the request is found to be valid.

The tables the models represent utilize a version system, which stores previous values for the column and allow for comparision against new versions being sent in the requests.

This is useful for applications with offline functionality because offline applications can queue up events to send when network connectivity is available, and the model can be updated in response to the queued events.

### Requests
When making an event `POST` request, the controller expects the request body to be an array of event objects in this format:
```json
{
  "data": [
    {
      "command": "{model}:{action}",
      "model_uuid": "{uuid}",
      "value": {
        "uuid": "{uuid}",
        "{column}": "{value}",
        "{column_version}": "{uuid}",
      }
    }
  ]
}
```

The `{model}` must be a `command` from one of the registered models from the `events.php` file.

The `{action}` must be one of the elements in the `protected $commands` property of the registered model.

The `{column}` must be one of the columns of the model. It cannot be one of the **version** columns.

The `{column_version}` must be a `uuid`. 

An example:
```json
{
  "data": [
    {
      "command": "carrier:create",
      "model_uuid": "0d001765-6b5d-4794-8bbc-97f2ca79e379",
      "value": {
        "uuid": "0d001765-6b5d-4794-8bbc-97f2ca79e379",
        "name": "John's Trucking",
        "name_version": "01a5732d-86f2-4172-8c80-564b72b79290"
      }
    },
    {
      "command": "commodity:create",
      "model_uuid": "01a5732d-86f2-4172-8c80-564b72b79290",
      "value": {
        "uuid": "01a5732d-86f2-4172-8c80-564b72b79290",
        "name": "Soft Red Winter Wheat",
        "name_version": "ca7245e8-4362-3b8a-86be-63a7ad5d7721",
        "upper_limit": 13428,
        "upper_limit_version": "5b3e6639-42f8-3577-8807-6321094dd9e6",
        "lower_limit": 2801,
        "lower_limit_version": "e555219b-7c0b-3dd0-abc2-65f6dc46ad17",
        "bushel_conversion": 14,
        "bushel_conversion_version": "0166aa86-aa43-3693-ac7a-249773750644"
      }
    },
```

### New Migrations

Each new migration should define each of its columns, along with a column for the version. 

For example, if a migration is defining a column `column`, a `column_version` column should also be defined:
```php
$table->string('column')->nullable();
$table->uuid('column_version')->nullable();
```

It is important that each **version** column be defined as a `uuid`, and that each column be `nullable`.

### Models

#### **EventModel**
abstract class **EventModel** extends **UUIDModel** implements **IUserFilterable**

##### Properties
*protected* `array` **$commands** - An array of the commands that can be performed on the model. Each extending class can define its own list of allowed commands.

*protected* `JsonResouce` **$resource** - The JsonResource used to transform the data. 

##### Methods
 Return | Signature | Description
 --- | --- | ---
 `array` | `getCommands()` | Return the array of commands available to the model
 `JsonResource` | `format()` | Returns the formatted resource
 `LengthAwarePaginator` | `getUpdated(int $lastUpdate, int $perPage)` | Returns a `LengthAwarePaginator` of updated models filtered by the `$lastUpdate` unix timestamp integer and the `$perPage` limit integer.
 `void` | `createModel(array $event)` | Create a new instance of the model. Extending classes can override.
 `void` | `updateModel(array $event)` | Update a model property and its version. Extending classes can override.
 `void` | `deleteModel(array $event)` | Delete a model instance. Extending classes can override.
 `boolean` | `isCurrent(string $name, string $version)` | Check if the version of the model's column matches the `$version` being passed in. 

#### **Event**
class Event extends UUIDModel

##### Properties
*protected* `array` **$fillable** - The mass assignable fields for the model.

##### Methods
Return | Signature | Description
--- | --- | ---
`static void` | `boot()` | Sets the `uuid` and `user_uuid` columns of the model when creating the model.

#### **FailedEvent**
class FailedEvent extends UUIDModel

##### Properties
*protected* `array` **$fillable** - The mass assignable fields for the model.

##### Methods
Return | Signature | Description
--- | --- | ---
`static void` | `boot()` | Sets the `uuid` and `user_uuid` columns of the model when creating the model.

### Repository

#### **EventRepository**
class EventRepository

##### Methods
Return | Signature | Description
--- | --- | ---
`Event` | `create(array $event)` | Create a new `Event` model

#### **FailedEventRepository**
class FailedEventRepository

##### Methods
Return | Signature | Description
--- | --- | ---
`FailedEvent` | `create(array $event, string $reason)` | Create a new `FailedEvent` model
`void` | `createAll(array $errors)` | Create a new `FailedEvent` model for each error in the array

### Service

#### **EventService**
class **EventService** 

##### Properties
*private* `EventRepository` **$eventRepository** - The repository that handles the database operations for the `Events` table.

*private* `FailedEventRepository` **$failedEventRepository** - The repository that handles the failed events.

##### Methods
Return | Signature | Description
--- | --- | ---
`void` | `__construct(EventRepository $eventRepository)` | The `EventService` constructor. The private `$eventRepository` property is set via the injection of the `EventRepository`.
`array` | `getNew(string $uri, int $lastUpdate, int $pageSize)` | Returns an associative array containing a collection of newly updated models, determined from the `$uri` and pagination data.
`Collection` | `handlePost(Request $request)` | Partition out the valid event data from the the `$request`, then iterate over the events, determine the model and operation from the event, and finally perform the operation on the model.
`void` | `create(EventModel $model, array $event)` | Validates the `$event`, then creates a new `Event` record and a new instance of the `$model`.
`void` | `update(EventModel $model, array $event)` | Validates the `$event`, then checks the version of the column being updated. If a newer version already exists, an exception is thrown. If the version is valid, a new `Event` record is created, then the `$model` is updated.
`void` | `delete(EventModel $model, array $event)` | Deletes the `$model`.

### Controller

#### **EventController**
class EventsController extends Controller

##### Methods
Return | Signature | Description
--- | --- | ---
`Response` | `getEvents(string $uri, PaginationRequest $request, EventService $eventService)` | Return a response containing new events for the model, determined from the `$uri`, and pagination data.
`Response` | `store(Request $request, EventService $eventService)` | Return a response after events have been processed, which may contain the error messages of failed events.

### Utilities

#### **EventHelper**
class EventHelper

##### Methods
Return | Signature | Description
--- | --- | ---
`static EventModel` | `getModelByUri(string $uri)` | Retrieve a model by passing the `uri` from the `Request`.
`static EventModel` | `getModelByType(string $modelType)` | Retrieve a model by passing the `type`.
`static string` | `getCommandForModel(Model $model)` | Retrieve the command from the config by passing in an instance of a `Model`.

#### **EventBuilder**
class EventBuilder

##### Methods
Return | Signature | Description
--- | --- | ---
`static array` | `make(string $model, EventModel $data, JsonResource $formatter = null)` | Build an event associative array to be encoded in JSON and sent in a `Response`.
`static mixed` | `format(mixed $value, mixed $formatter)` | Format the event array `value` property.

#### **Pagination**
class Pagination

##### Methods
Return | Signature | Description
--- | --- | ---
`static array` | `build(LengthAwarePaginator $paginator)` | Build an array of pagination information, often for including in a `Response`. 